/*
 * filter_input.h
 * Copyright 2014 Daniel (dmilith) Dettlaff
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions, and the following disclaimer in the documentation
 *    provided with the distribution.
 *
 * This software is provided "as is" and without any warranty, express or
 * implied. In no event shall the authors be liable for any damages arising from
 * the use of this software.
 */

#ifndef FILTER_INPUT_WIDGET_H
#define FILTER_INPUT_WIDGET_H

#include <QLineEdit>

class FilterInput : public QLineEdit
{
public:
    FilterInput (QWidget * parent = nullptr);

protected:
    virtual void keyPressEvent (QKeyEvent * e); /* override default handler */
};

#endif
